import os
import json
import csv
from pathlib import Path

import openpyxl

from smooth_criminal.core import auto_boost
from smooth_criminal.memory import export_execution_history, LOG_PATH

@auto_boost()
def export_test_func():
    return sum(i for i in range(1000))

def test_export_to_multiple_formats(tmp_path):
    # Ejecutar función varias veces para llenar historial
    for _ in range(2):
        export_test_func()

    # Rutas de exportación temporales
    csv_path = tmp_path / "export.csv"
    json_path = tmp_path / "export.json"
    xlsx_path = tmp_path / "export.xlsx"
    md_path = tmp_path / "export.md"

    # Exportar a CSV
    result_csv = export_execution_history(csv_path, format="csv")
    assert result_csv
    assert csv_path.exists()

    # Comprobar contenido del CSV
    with open(csv_path, newline='', encoding='utf-8') as f:
        reader = csv.DictReader(f)
        rows = list(reader)
        assert len(rows) >= 1
        assert "function" in rows[0]
        assert rows[0]["function"] == "export_test_func"

    # Exportar a JSON
    result_json = export_execution_history(json_path, format="json")
    assert result_json
    assert json_path.exists()

    with open(json_path, encoding='utf-8') as f:
        data = json.load(f)
        assert isinstance(data, list)
        assert any(entry["function"] == "export_test_func" for entry in data)

    # Exportar a XLSX
    result_xlsx = export_execution_history(xlsx_path, format="xlsx")
    assert result_xlsx
    assert xlsx_path.exists()
    wb = openpyxl.load_workbook(xlsx_path)
    ws = wb.active
    rows = list(ws.iter_rows(values_only=True))
    assert rows[0][0] == "function"
    assert any(r[0] == "export_test_func" for r in rows[1:])

    # Exportar a Markdown
    result_md = export_execution_history(md_path, format="md")
    assert result_md
    assert md_path.exists()
    with open(md_path, encoding="utf-8") as f:
        content = f.read()
        assert "export_test_func" in content
        assert content.startswith("| function |")



import os
import tempfile
import matplotlib.pyplot as plt
from smooth_criminal.memory import get_execution_history
from smooth_criminal.core import auto_boost


@auto_boost()
def func_to_graph():
    return sum(i for i in range(1000))

def test_graph_generation():
    # Ejecutar varias veces para crear historial
    for _ in range(5):
        func_to_graph()

    history = get_execution_history()
    func_name = "func_to_graph"
    times = [entry["duration"] for entry in history if entry["function"] == func_name]

    assert len(times) >= 3, "No hay suficientes datos para graficar."

    # Generar gráfico y guardarlo temporalmente
    fig, ax = plt.subplots()
    ax.plot(times, marker='o')
    ax.set_title(f"Historial de tiempos: {func_name}")
    ax.set_xlabel("Ejecución")
    ax.set_ylabel("Duración (s)")

    with tempfile.NamedTemporaryFile(delete=False, suffix=".png") as tmp:
        fig.savefig(tmp.name)
        path = tmp.name

    plt.close(fig)

    # Verificar que el archivo existe y tiene tamaño
    assert os.path.exists(path)
    assert os.path.getsize(path) > 0

    # Limpieza
    os.remove(path)



import argparse
import importlib
import importlib.util
import inspect
import logging
import os
import json

from rich.console import Console
from rich.logging import RichHandler
from rich.table import Table

from smooth_criminal.analizer import analyze_ast
from smooth_criminal.benchmark import benchmark_jam
from smooth_criminal.dashboard import render_dashboard
from smooth_criminal.memory import (
    suggest_boost,
    clear_execution_history,
    export_execution_history,
    score_function,
)
from smooth_criminal.core import play_mj_effect, set_mj_mode


log_level = os.getenv("LOG_LEVEL", "INFO").upper()
numeric_level = getattr(logging, log_level, logging.INFO)

logging.basicConfig(
    level=numeric_level,
    format="%(message)s",
    handlers=[RichHandler(rich_tracebacks=True, markup=True)],
    force=True,
)

logger = logging.getLogger("SmoothCriminal")

def main():
    parser = argparse.ArgumentParser(description="Smooth Criminal CLI")
    parser.add_argument(
        "--mj-mode",
        action="store_true",
        help="Activa efectos MJ cuando hay mejoras de rendimiento",
    )
    subparsers = parser.add_subparsers(dest="command")

    # Comando 'analyze'
    analyze_parser = subparsers.add_parser("analyze", help="Analiza un archivo Python en busca de optimizaciones.")
    analyze_parser.add_argument("filepath", help="Ruta del script a analizar")

    # Comando 'suggest'
    suggest_parser = subparsers.add_parser("suggest", help="Sugiere el mejor decorador para una función registrada.")
    suggest_parser.add_argument("func_name", help="Nombre de la función a consultar")

    # Comando 'dashboard'
    subparsers.add_parser("dashboard", help="Muestra un resumen del historial de optimizaciones.")

    # Comando 'clean'
    subparsers.add_parser("clean", help="Elimina el historial de ejecuciones registrado.")

    # Comando 'export'
    export_parser = subparsers.add_parser(
        "export", help="Exporta el historial como archivo CSV, JSON, XLSX o Markdown."
    )
    export_parser.add_argument("filepath", help="Ruta del archivo de salida")
    export_parser.add_argument(
        "--format",
        choices=["csv", "json", "xlsx", "md"],
        default="csv",
        help="Formato de exportación",
    )

    # Comando 'score'
    score_parser = subparsers.add_parser("score", help="Muestra una puntuación de optimización para una función.")
    score_parser.add_argument("func_name", help="Nombre de la función a evaluar")

    # Comando 'jam-test'
    jam_parser = subparsers.add_parser(
        "jam-test", help="Benchmark de backends de jam para una función"
    )
    jam_parser.add_argument(
        "func_path", help="Ruta a la función en formato modulo:funcion"
    )
    jam_parser.add_argument(
        "--workers", type=int, default=1, help="Número de workers a utilizar"
    )
    jam_parser.add_argument(
        "--reps", type=int, default=1, help="Número de repeticiones del benchmark"
    )
    jam_parser.add_argument(
        "--silent",
        action="store_true",
        help="Muestra solo el resultado en formato JSON",
    )

    args = parser.parse_args()
    set_mj_mode(args.mj_mode)
    if not getattr(args, "silent", False):
        show_intro()

    if args.command == "analyze":
        analyze_file(args.filepath)
    elif args.command == "suggest":
        handle_suggestion(args.func_name)
    elif args.command == "dashboard":
        render_dashboard()
    elif args.command == "clean":
        handle_clean()
    elif args.command == "export":
        handle_export(args.filepath, args.format)
    elif args.command == "score":
        handle_score(args.func_name)
    elif args.command == "jam-test":
        handle_jam_test(
            args.func_path, args.workers, args.reps, args.silent, args.mj_mode
        )

    else:
        logger.warning(
            "[bold red]Annie, are you OK?[/bold red] "
            "Use '[green]smooth-criminal analyze <file.py>[/green]', "
            "'[green]smooth-criminal suggest <func>[/green]', "
            "'[green]smooth-criminal dashboard[/green]' or "
            "'[green]smooth-criminal export <file> --format csv/json/xlsx/md[/green]'."
        )

def show_intro():
    logger.info(
        "[bold magenta]Smooth Criminal[/bold magenta]\n"
        "[dim]By Adolfo – Powered by Python + MJ energy 🕺[/dim]"
    )

def analyze_file(filepath):
    logger.info(
        f"\n🎤 [cyan]Analyzing [bold]{filepath}[/bold] for optimization opportunities...[/cyan]\n"
    )

    spec = importlib.util.spec_from_file_location("target_module", filepath)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    functions = [getattr(module, name) for name in dir(module)
                 if inspect.isfunction(getattr(module, name)) and getattr(module, name).__module__ == module.__name__]

    if not functions:
        logger.warning("[yellow]No se encontraron funciones para analizar.[/yellow]")
        return

    for func in functions:
        logger.info(
            f"\n🔎 [bold blue]Analyzing function:[/bold blue] [white]{func.__name__}[/white]"
        )
        analyze_ast(func)

def handle_suggestion(func_name):
    logger.info(
        f"\n🧠 [bold cyan]Consulting memory for:[/bold cyan] [white]{func_name}[/white]\n"
    )
    suggestion = suggest_boost(func_name)
    logger.info(suggestion)

def handle_clean():
    if clear_execution_history():
        logger.info("[green]Historial de ejecuciones borrado exitosamente.[/green]")
    else:
        logger.warning("[yellow]No se encontró historial para borrar.[/yellow]")

def handle_export(filepath, format):
    success = export_execution_history(filepath, format)
    if success:
        logger.info(
            f"[green]Historial exportado a [bold]{filepath}[/bold] como {format.upper()}.[/green]"
        )
    else:
        logger.warning("[yellow]No hay historial para exportar.[/yellow]")

def handle_score(func_name):
    score, summary = score_function(func_name)
    if score is None:
        logger.warning(f"[yellow]{summary}[/yellow]")
    else:
        logger.info(summary)
        color = "green" if score >= 80 else "yellow" if score >= 50 else "red"
        logger.info(f"[bold {color}]Optimization Score: {score}/100[/bold {color}]")


def handle_jam_test(
    func_path: str, workers: int, reps: int, silent: bool, mj_mode: bool
) -> None:
    module_name, func_name = func_path.split(":", 1)
    module = importlib.import_module(module_name)
    func = getattr(module, func_name)

    backends = ["thread", "process", "async"]
    args = list(range(workers))
    if silent:
        logging.disable(logging.CRITICAL)

    totals = {b: 0.0 for b in backends}
    for _ in range(reps):
        result = benchmark_jam(func, args, backends)
        if not all(m.get("success") for m in result["metrics"]):
            logger.error("[red]One or more backends failed during benchmarking.[/red]")
            if silent:
                logging.disable(logging.NOTSET)
            return
        for metric in result["metrics"]:
            totals[metric["backend"]] += metric["duration"]

    averages = {b: totals[b] / reps for b in backends}
    fastest = min(averages, key=averages.get)
    if silent:
        print(json.dumps({"averages": averages, "fastest": fastest}))
        logging.disable(logging.NOTSET)
        return

    max_duration = max(averages.values()) or 1.0
    min_duration = min(averages.values()) or 0.0
    improvement = ((max_duration - min_duration) / max_duration) * 100
    play_mj_effect(improvement, mj_mode)

    console = Console()
    table = Table(title="Resultados jam-test")
    table.add_column("Backend")
    table.add_column("Promedio (s)", justify="right")
    table.add_column("Comparativa")

    for backend in backends:
        duration = averages[backend]
        bar_length = int((duration / max_duration) * 40)
        bar = "█" * bar_length
        table.add_row(backend, f"{duration:.4f}", bar)

    console.print(table)
    logger.info("🎶 Just jammin' through those CPU cores! 🧠🕺")


if __name__ == "__main__":
    main()



import asyncio
import flet as ft

def info_panel(text: str, color="blue") -> ft.Container:
    return ft.Container(
        content=ft.Text(text, color=color, size=16),
        padding=10,
        bgcolor=ft.colors.SURFACE_VARIANT,
        border_radius=10
    )

def function_table() -> ft.DataTable:
    return ft.DataTable(
        columns=[
            ft.DataColumn(ft.Text("Function")),
            ft.DataColumn(ft.Text("Decorator(s)")),
            ft.DataColumn(ft.Text("Runs")),
            ft.DataColumn(ft.Text("Avg Time (s)")),
            ft.DataColumn(ft.Text("Score")),
        ],
        rows=[]
    )

def action_buttons(refresh_fn, clear_fn, export_fn, graph_fn) -> ft.Row:
    return ft.Row([
        ft.ElevatedButton("🔄 Refresh", on_click=refresh_fn, icon=ft.Icons.REFRESH),
        ft.ElevatedButton("🧼 Limpiar historial", on_click=clear_fn, icon=ft.Icons.DELETE),
        ft.ElevatedButton("💾 Exportar CSV", on_click=lambda e: export_fn("csv"), icon=ft.Icons.DOWNLOAD),
        ft.ElevatedButton("💾 Exportar JSON", on_click=lambda e: export_fn("json"), icon=ft.Icons.DOWNLOAD),
        ft.ElevatedButton("💾 Exportar XLSX", on_click=lambda e: export_fn("xlsx"), icon=ft.Icons.DOWNLOAD),
        ft.ElevatedButton("💾 Exportar MD", on_click=lambda e: export_fn("md"), icon=ft.Icons.DOWNLOAD),
        ft.ElevatedButton("📈 Ver gráfico", on_click=graph_fn, icon=ft.Icons.INSERT_CHART)
    ], spacing=15)


def moonwalk_animation(page: ft.Page, duration: int = 800) -> None:
    """Desplaza un ícono de MJ por la pantalla y limpia al finalizar."""

    dancer = ft.Text("🕺", size=40)
    anim = ft.AnimatedContainer(
        content=dancer,
        width=40,
        height=40,
        left=-50,
        top=page.height / 2,
        animate_position=ft.animation.Animation(duration, "ease"),
    )
    page.overlay.append(anim)
    page.update()

    anim.left = page.width
    anim.update()

    async def _cleanup():
        await asyncio.sleep(duration / 1000)
        page.overlay.remove(anim)
        await page.update_async()

    page.run_task(_cleanup)



from smooth_criminal.core import auto_boost
from smooth_criminal.dashboard import render_dashboard
from smooth_criminal.memory import get_execution_history
from io import StringIO
from rich.console import Console
import sys

@auto_boost()
def dashboard_test_func():
    return sum(i for i in range(3000))

def test_render_dashboard_output(capsys):
    # Ejecutar la función varias veces para llenar historial
    for _ in range(3):
        dashboard_test_func()

    # Redirigir salida a buffer
    console = Console(file=StringIO())
    original_stdout = sys.stdout
    sys.stdout = console.file

    # Ejecutar dashboard
    render_dashboard()

    # Restaurar stdout y obtener contenido
    sys.stdout = original_stdout
    output = console.file.getvalue()

    # Verificamos que se imprimió la función
    assert "dashboard_test_func" in output
    assert "@smooth" in output or "@jam" in output
    assert "Avg Time" in output or "avg" in output.lower()



import os
from pathlib import Path
from smooth_criminal.core import auto_boost
from smooth_criminal.memory import suggest_boost

LOG_PATH = Path.home() / ".smooth_criminal_log.json"

@auto_boost()
def test_func_to_suggest():
    return sum(i for i in range(500))

def test_suggest_boost_recommendation():
    # Limpiar log antes del test si existe
    if LOG_PATH.exists():
        LOG_PATH.unlink()

    # Ejecutar varias veces para generar datos
    for _ in range(3):
        test_func_to_suggest()

    # Obtener sugerencia
    suggestion = suggest_boost("test_func_to_suggest")

    # Validar que sugiere algo coherente
    assert "use" in suggestion
    assert "@smooth" in suggestion or "@jam" in suggestion



from smooth_criminal.benchmark import benchmark_jam, detect_fastest_backend


def cube(x: int) -> int:
    return x ** 3


def test_benchmark_generates_metrics_for_all_backends():
    args = [1, 2, 3]
    result = benchmark_jam(cube, args, ["thread", "process", "async"])
    backends = {m["backend"] for m in result["metrics"]}
    assert backends == {"thread", "process", "async"}
    assert result["fastest"] in backends


def test_detect_fastest_backend_returns_valid_backend():
    best = detect_fastest_backend(cube, [1, 2, 3], ["thread", "process", "async"])
    assert best in {"thread", "process", "async"}



import flet as ft
from smooth_criminal.memory import (
    get_execution_history,
    clear_execution_history,
    export_execution_history,
    build_summary,
    calcular_score,
)
from smooth_criminal.flet_app.components import (
    info_panel,
    function_table,
    action_buttons,
    moonwalk_animation,
)
from smooth_criminal.flet_app.utils import formatear_tiempo, export_filename

def main_view(page: ft.Page):
    table = function_table()
    msg = ft.Text()

    def refresh(_=None):
        history = get_execution_history()
        summary = build_summary(history)

        table.rows.clear()
        for fn, data in summary.items():
            avg = sum(data["durations"]) / len(data["durations"])
            score = calcular_score(data["durations"], data["decorators"])
            table.rows.append(ft.DataRow(cells=[
                ft.DataCell(ft.Text(fn)),
                ft.DataCell(ft.Text(", ".join(sorted(data["decorators"])))),
                ft.DataCell(ft.Text(str(len(data["durations"])))),
                ft.DataCell(ft.Text(formatear_tiempo(avg))),
                ft.DataCell(ft.Text(f"{score}/100")),
            ]))
        page.update()

    def clear(_):
        if clear_execution_history():
            msg.value = "Historial borrado."
            table.rows.clear()
        else:
            msg.value = "No hay historial para borrar."
        page.update()

    def export(fmt: str):
        filepath = export_filename(ext=fmt)
        if export_execution_history(filepath, format=fmt):
            msg.value = f"Historial exportado a {filepath}"
        else:
            msg.value = "No hay historial para exportar."
        page.update()

    def trigger_moonwalk(e):
        moonwalk_animation(page)
        moonwalk_switch.value = False
        page.update()

    btns = action_buttons(refresh, clear, export, lambda e: None)
    moonwalk_switch = ft.Switch(label="Moonwalk", on_change=trigger_moonwalk)
    btns.controls.append(moonwalk_switch)

    page.add(
        ft.Text("🎩 Smooth Criminal Dashboard", size=28, weight="bold", color="purple"),
        btns,
        msg,
        table
    )

    refresh()



"""Módulo de persistencia configurable para Smooth Criminal.

Este archivo define una interfaz ``StorageBackend`` con tres
implementaciones disponibles: ``JsonBackend``, ``SQLiteBackend`` y
``TinyDBBackend``.  El backend se selecciona mediante la variable de
entorno ``SMOOTH_CRIMINAL_STORAGE`` que puede tomar los valores
``json`` (por defecto), ``sqlite`` o ``tinydb``.

Las funciones públicas del módulo delegan su comportamiento en el
backend elegido manteniendo la API original para el resto del
proyecto.
"""

from __future__ import annotations

import csv
import json
import os
import statistics
from abc import ABC, abstractmethod
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Set


def calcular_score(durations: List[float], decorators: Set[str]) -> int:
    """Calcula una puntuación de optimización basada en duración y decoradores."""
    if not durations:
        return 0

    avg = statistics.mean(durations)
    stddev = statistics.stdev(durations) if len(durations) > 1 else 0.0

    score = 100
    if "@smooth" not in decorators and "@jam" not in decorators:
        score -= 20
    if avg > 0.01:
        score -= min((avg * 1000), 20)
    if stddev > 0.005:
        score -= 10

    return max(0, round(score))


class StorageBackend(ABC):
    """Interfaz base para los distintos métodos de almacenamiento."""

    #: ruta por defecto utilizada por el backend
    path: Path

    @abstractmethod
    def log_execution_stats(
        self, func_name: str, input_type, decorator_used: str, duration: float
    ) -> None:
        """Guarda un registro de ejecución."""

    @abstractmethod
    def get_execution_history(self, func_name: Optional[str] = None) -> List[Dict]:
        """Obtiene el historial de ejecuciones."""

    def clear_execution_history(self) -> bool:
        """Limpia por completo el historial basado en :attr:`self.path`."""
        if self.path.exists():
            self.path.unlink()
            return True
        return False

    def export_execution_history(self, filepath, format: str = "csv") -> bool:
        """Exporta el historial a CSV, JSON, XLSX o Markdown."""
        data = self.get_execution_history()
        if not data:
            return False

        try:
            data.sort(key=lambda x: x.get("timestamp", ""), reverse=True)
        except Exception:
            pass

        format = format.lower()
        keys = ["function", "input_type", "decorator", "duration", "timestamp"]
        if format == "json":
            with open(filepath, "w", encoding="utf-8") as f:
                json.dump(data, f, indent=2)
        elif format == "csv":
            with open(filepath, "w", newline="", encoding="utf-8") as f:
                writer = csv.DictWriter(f, fieldnames=keys)
                writer.writeheader()
                writer.writerows(data)
        elif format == "xlsx":
            try:
                from openpyxl import Workbook
            except Exception as exc:  # pragma: no cover - se captura en pruebas
                raise RuntimeError(
                    "openpyxl es requerido para exportar a XLSX"
                ) from exc

            wb = Workbook()
            ws = wb.active
            ws.append(keys)
            for row in data:
                ws.append([row.get(k, "") for k in keys])
            wb.save(filepath)
        elif format == "md":
            with open(filepath, "w", encoding="utf-8") as f:
                header = "| " + " | ".join(keys) + " |\n"
                separator = "|" + " --- |" * len(keys) + "\n"
                f.write(header)
                f.write(separator)
                for row in data:
                    line = "| " + " | ".join(str(row.get(k, "")) for k in keys) + " |\n"
                    f.write(line)
        else:
            raise ValueError("Formato no soportado: usa 'csv', 'json', 'xlsx' o 'md'.")

        return True

    def score_function(self, func_name: str) -> Tuple[Optional[int], str]:
        """Calcula la puntuación de optimización para una función."""
        logs = self.get_execution_history(func_name)
        if not logs:
            return None, "No hay registros para esta función."

        times = [entry["duration"] for entry in logs]
        decorators = {entry["decorator"] for entry in logs}
        count = len(times)
        avg = statistics.mean(times)
        stddev = statistics.stdev(times) if count > 1 else 0.0

        score = calcular_score(times, decorators)

        summary = (
            f"🧠 Function: {func_name}\n"
            f"- Executions: {count}\n"
            f"- Avg time: {avg:.6f}s\n"
            f"- Std dev: {stddev:.6f}s\n"
            f"- Decorators: {', '.join(sorted(decorators))}\n"
        )

        return score, summary


# ---------------------------------------------------------------------------
# Implementaciones concretas


class JsonBackend(StorageBackend):
    """Persistencia basada en un archivo JSON."""

    path = Path.home() / ".smooth_criminal_log.json"

    def log_execution_stats(self, func_name, input_type, decorator_used, duration):
        log_entry = {
            "function": func_name,
            "input_type": str(input_type),
            "decorator": decorator_used,
            "duration": duration,
            "timestamp": datetime.utcnow().isoformat(),
        }

        logs: List[Dict] = []
        if self.path.exists():
            with open(self.path, "r", encoding="utf-8") as f:
                try:
                    logs = json.load(f)
                except json.JSONDecodeError:
                    logs = []

        logs.append(log_entry)
        with open(self.path, "w", encoding="utf-8") as f:
            json.dump(logs, f, indent=2)

    def get_execution_history(self, func_name: Optional[str] = None) -> List[Dict]:
        if not self.path.exists():
            return []

        with open(self.path, "r", encoding="utf-8") as f:
            try:
                logs = json.load(f)
            except json.JSONDecodeError:
                return []

        if func_name:
            logs = [entry for entry in logs if entry["function"] == func_name]
        return logs


class SQLiteBackend(StorageBackend):
    """Persistencia utilizando una base de datos SQLite."""

    path = Path.home() / ".smooth_criminal_log.sqlite"

    def _ensure_table(self, conn) -> None:
        conn.execute(
            """
            CREATE TABLE IF NOT EXISTS logs (
                function TEXT,
                input_type TEXT,
                decorator TEXT,
                duration REAL,
                timestamp TEXT
            )
            """
        )
        conn.commit()

    def log_execution_stats(self, func_name, input_type, decorator_used, duration):
        import sqlite3

        with sqlite3.connect(self.path) as conn:
            self._ensure_table(conn)
            conn.execute(
                "INSERT INTO logs VALUES (?, ?, ?, ?, ?)",
                (
                    func_name,
                    str(input_type),
                    decorator_used,
                    float(duration),
                    datetime.utcnow().isoformat(),
                ),
            )
            conn.commit()

    def get_execution_history(self, func_name: Optional[str] = None) -> List[Dict]:
        import sqlite3

        if not self.path.exists():
            return []

        with sqlite3.connect(self.path) as conn:
            self._ensure_table(conn)
            cursor = conn.cursor()
            if func_name:
                cursor.execute(
                    "SELECT function,input_type,decorator,duration,timestamp FROM logs WHERE function=?",
                    (func_name,),
                )
            else:
                cursor.execute(
                    "SELECT function,input_type,decorator,duration,timestamp FROM logs"
                )
            rows = cursor.fetchall()

        return [
            {
                "function": r[0],
                "input_type": r[1],
                "decorator": r[2],
                "duration": r[3],
                "timestamp": r[4],
            }
            for r in rows
        ]


class TinyDBBackend(StorageBackend):
    """Persistencia mediante TinyDB."""

    path = Path.home() / ".smooth_criminal_log.tinydb"

    def __init__(self) -> None:
        try:
            from tinydb import Query, TinyDB
        except Exception as exc:  # pragma: no cover - la importación es dinámica
            raise RuntimeError(
                "TinyDB no está instalado. Instálalo con el extra 'tinydb'."
            ) from exc

        self.TinyDB = TinyDB
        self.Query = Query

    def _open(self):
        return self.TinyDB(self.path)

    def log_execution_stats(self, func_name, input_type, decorator_used, duration):
        with self._open() as db:
            db.insert(
                {
                    "function": func_name,
                    "input_type": str(input_type),
                    "decorator": decorator_used,
                    "duration": duration,
                    "timestamp": datetime.utcnow().isoformat(),
                }
            )

    def get_execution_history(self, func_name: Optional[str] = None) -> List[Dict]:
        if not self.path.exists():
            return []

        with self._open() as db:
            if func_name:
                results = db.search(self.Query().function == func_name)
            else:
                results = db.all()
        return list(results)


# ---------------------------------------------------------------------------
# Selección dinámica del backend


def _select_backend() -> StorageBackend:
    name = os.getenv("SMOOTH_CRIMINAL_STORAGE", "json").lower()
    if name == "sqlite":
        return SQLiteBackend()
    if name == "tinydb":
        return TinyDBBackend()
    return JsonBackend()


_BACKEND: StorageBackend = _select_backend()

# Exponer la ruta del backend seleccionado para mantener compatibilidad con
# el código existente y las pruebas.
LOG_PATH = _BACKEND.path


def log_execution_stats(func_name, input_type, decorator_used, duration):
    """Delegación pública al backend activo."""
    _BACKEND.log_execution_stats(func_name, input_type, decorator_used, duration)


_ORIGINAL_GET_HISTORY = None


def get_execution_history(func_name=None):
    """Obtiene el historial usando el backend activo.

    Mantiene compatibilidad con monkeypatch en tests que importaron una
    referencia previa de la función.
    """

    current = globals().get("get_execution_history", _ORIGINAL_GET_HISTORY)
    if _ORIGINAL_GET_HISTORY is not None and current is not _ORIGINAL_GET_HISTORY:
        try:
            return current(func_name)
        except TypeError:  # pragma: no cover - compatibilidad
            return current()

    return _BACKEND.get_execution_history(func_name)


_ORIGINAL_GET_HISTORY = get_execution_history


def clear_execution_history():
    """Limpia el historial usando el backend activo."""
    return _BACKEND.clear_execution_history()


def export_execution_history(filepath, format="csv"):
    """Exporta el historial usando el backend activo."""
    return _BACKEND.export_execution_history(filepath, format=format)


def build_summary(logs: List[Dict]) -> Dict[str, Dict[str, object]]:
    """Agrupa un listado de logs por función.

    Parameters
    ----------
    logs:
        Secuencia de registros devueltos por :func:`get_execution_history`.

    Returns
    -------
    dict
        Diccionario cuyas claves son los nombres de función y los valores
        contienen dos entradas:

        ``durations``
            Lista con las duraciones registradas.
        ``decorators``
            Conjunto con los decoradores utilizados.
    """

    summary: Dict[str, Dict[str, object]] = {}
    for entry in logs:
        fn = entry.get("function")
        if fn is None:
            continue
        data = summary.setdefault(fn, {"durations": [], "decorators": set()})
        if "duration" in entry:
            data["durations"].append(entry["duration"])
        if "decorator" in entry:
            data["decorators"].add(entry["decorator"])
    return summary


def score_function(func_name):
    """Calcula la puntuación delegando en el backend activo."""
    return _BACKEND.score_function(func_name)


def suggest_boost(func_name):
    """Función auxiliar original; se mantiene sin cambios."""
    logs = get_execution_history(func_name)
    if not logs:
        return f"No data found for function '{func_name}'."

    decor_stats = {}
    for entry in logs:
        decor_stats.setdefault(entry["decorator"], []).append(entry["duration"])

    avg_times = {decor: sum(times) / len(times) for decor, times in decor_stats.items()}
    best_decor = min(avg_times, key=avg_times.get)
    return (
        f"🧠 Suggestion for '{func_name}': use [bold green]{best_decor}[/bold green] "
        f"(avg {avg_times[best_decor]:.6f}s)"
    )


# Mantener compatibilidad con el sistema de parcheo utilizado por algunas
# pruebas; se almacena una referencia a la función original.
_ORIGINAL_GET_HISTORY = get_execution_history




import subprocess
import sys
import json
from pathlib import Path


def test_jam_test_cli_output():
    repo_root = Path(__file__).resolve().parent.parent
    cmd = [
        sys.executable,
        "-m",
        "smooth_criminal.cli",
        "jam-test",
        "tests.sample_funcs:compute",
        "--workers",
        "2",
        "--reps",
        "1",
    ]
    result = subprocess.run(cmd, capture_output=True, text=True, cwd=repo_root)
    output = result.stdout + result.stderr
    assert "thread" in output
    assert "process" in output
    assert "async" in output
    assert result.returncode == 0


def test_jam_test_cli_silent_json():
    repo_root = Path(__file__).resolve().parent.parent
    cmd = [
        sys.executable,
        "-m",
        "smooth_criminal.cli",
        "jam-test",
        "tests.sample_funcs:compute",
        "--workers",
        "2",
        "--reps",
        "1",
        "--silent",
    ]
    result = subprocess.run(cmd, capture_output=True, text=True, cwd=repo_root)
    assert result.stderr == ""
    data = json.loads(result.stdout)
    assert set(data["averages"].keys()) == {"thread", "process", "async"}
    assert result.returncode == 0


def test_jam_test_cli_message_non_silent():
    repo_root = Path(__file__).resolve().parent.parent
    cmd = [
        sys.executable,
        "-m",
        "smooth_criminal.cli",
        "jam-test",
        "tests.sample_funcs:compute",
        "--workers",
        "2",
        "--reps",
        "1",
    ]
    result = subprocess.run(cmd, capture_output=True, text=True, cwd=repo_root)
    output = result.stdout + result.stderr
    assert "🎶 Just jammin' through those CPU cores! 🧠🕺" in output
    assert result.returncode == 0




import pytest
pytest.skip("Componentes Flet no disponibles", allow_module_level=True)

from smooth_criminal.flet_app.components import (
    function_table,
    action_buttons,
    info_panel
)
import flet as ft

def test_function_table_structure():
    table = function_table()
    assert isinstance(table, ft.DataTable)
    assert len(table.columns) == 5
    expected_headers = ["Function", "Decorator(s)", "Runs", "Avg Time (s)", "Score"]
    headers = [col.label.value for col in table.columns]
    assert headers == expected_headers

def test_action_buttons_creation():
    def dummy(*args): pass
    row = action_buttons(dummy, dummy, dummy, dummy)
    assert isinstance(row, ft.Row)
    assert len(row.controls) == 4
    labels = [btn.text for btn in row.controls]
    assert "Refresh" in labels[0]
    assert "Limpiar" in labels[1]
    assert "Exportar" in labels[2]
    assert "Gráfico" in labels[3]

def test_info_panel_display():
    panel = info_panel("Hola mundo", color="green")
    assert isinstance(panel, ft.Container)
    assert isinstance(panel.content, ft.Text)
    assert panel.content.value == "Hola mundo"
    assert panel.content.color == "green"



from smooth_criminal.core import auto_boost
from smooth_criminal.memory import score_function, clear_execution_history

@auto_boost()
def score_test_func():
    return sum(i for i in range(1000))

def test_score_function_evaluation():
    # Limpiar historial antes
    clear_execution_history()

    # Generar registros
    for _ in range(5):
        score_test_func()

    # Obtener puntuación
    score, summary = score_function("score_test_func")

    assert isinstance(score, int)
    assert 0 <= score <= 100
    assert "score_test_func" in summary
    assert "- Executions:" in summary
    assert "- Avg time:" in summary
    assert "- Decorators:" in summary



import pytest
from smooth_criminal.core import bad

@bad(parallel=False)
def sum_sqrt(n):
    total = 0.0
    for i in range(1, n + 1):
        total += i ** 0.5
    return round(total, 2)

def test_bad_aggressive_optimization():
    result = sum_sqrt(1000)
    expected = round(sum(i ** 0.5 for i in range(1, 1001)), 2)
    assert abs(result - expected) < 1e-6

def test_bad_exception():
    def fail(x, y):
        raise ValueError("boom")
    wrapped = bad(parallel=False)(fail)
    with pytest.raises(ValueError):
        wrapped(1, 2)



import pytest
from smooth_criminal.memory import build_summary


def test_build_summary_basic():
    logs = [
        {"function": "foo", "duration": 0.1, "decorator": "@smooth"},
        {"function": "foo", "duration": 0.2, "decorator": "@smooth"},
        {"function": "bar", "duration": 0.3, "decorator": "@jam"},
    ]
    summary = build_summary(logs)
    assert set(summary.keys()) == {"foo", "bar"}
    assert summary["foo"]["decorators"] == {"@smooth"}
    assert summary["bar"]["durations"] == [0.3]


def test_build_summary_empty():
    assert build_summary([]) == {}



import json
import pytest
from pathlib import Path
from smooth_criminal.core import auto_boost
from smooth_criminal.memory import get_execution_history

LOG_PATH = Path.home() / ".smooth_criminal_log.json"


@auto_boost()
def boosted_function():
    return sum(i for i in range(1000))


@auto_boost()
def square(x):
    return x * x


def fallback(_):
    return -1


@auto_boost(fallback=fallback)
def failing(x):
    raise ValueError("boom")


def test_auto_boost_memory_logs():
    if LOG_PATH.exists():
        LOG_PATH.unlink()
    boosted_function()
    assert LOG_PATH.exists()
    history = get_execution_history("boosted_function")
    assert len(history) > 0
    entry = history[-1]
    assert entry["function"] == "boosted_function"
    assert entry["decorator"] in ("@smooth", "@jam", "none")
    assert isinstance(entry["duration"], float)
    assert entry["duration"] > 0


def test_auto_boost_normal(numbers):
    result = square(numbers[:3])
    assert sorted(result) == [1, 4, 9]


def test_auto_boost_fallback(numbers):
    result = failing(numbers[:3])
    assert result == [-1, -1, -1]



from smooth_criminal.memory import calcular_score
from smooth_criminal.flet_app.utils import (
    formatear_tiempo,
    export_filename
)

def test_calcular_score_basic():
    durations = [0.001, 0.002, 0.0015]
    decorators = {"@smooth"}
    score = calcular_score(durations, decorators)
    assert isinstance(score, int)
    assert 80 <= score <= 100

def test_calcular_score_penalized():
    durations = [0.02, 0.03, 0.025]
    decorators = {"@none"}
    score = calcular_score(durations, decorators)
    assert score < 80

def test_formatear_tiempo():
    tiempo = 0.002345678
    formatted = formatear_tiempo(tiempo)
    assert formatted.endswith("s")
    assert formatted.startswith("0.002345")

def test_export_filename_format():
    name = export_filename()
    assert name.startswith("smooth_export_")
    assert name.endswith(".csv")
    assert len(name) > 20



import logging
import os

from rich.logging import RichHandler

from smooth_criminal.core import jam


log_level = os.getenv("LOG_LEVEL", "INFO").upper()
numeric_level = getattr(logging, log_level, logging.INFO)

logging.basicConfig(
    level=numeric_level,
    format="%(message)s",
    handlers=[RichHandler(rich_tracebacks=True, markup=True)],
    force=True,
)

logger = logging.getLogger("SmoothCriminal")

@jam(workers=4, backend="thread")  # Cambia backend a "process" o "async" según necesidad
def double(x):
    return x * 2

if __name__ == "__main__":
    numbers = list(range(10_000))
    result = double(numbers)
    logger.info(f"Processed {len(result)} numbers.")



import flet as ft
from smooth_criminal.memory import (
    get_execution_history,
    score_function,
    clear_execution_history,
    export_execution_history,
    build_summary,
)
from datetime import datetime

def mostrar_grafico(e, page, dropdown_func, msg):
    import matplotlib.pyplot as plt
    import tempfile
    from flet import Image

    history = get_execution_history()
    if not history:
        msg.value = "⚠️ No hay datos para graficar."
        page.update()
        return

    func = dropdown_func.value if dropdown_func.value else history[0]["function"]
    times = [entry["duration"] for entry in history if entry["function"] == func]

    if not times:
        msg.value = "⚠️ No hay datos para graficar."
        page.update()
        return

    fig, ax = plt.subplots()
    ax.plot(times, marker='o')
    ax.set_title(f"Historial de tiempos: {func}")
    ax.set_xlabel("Ejecución")
    ax.set_ylabel("Duración (s)")

    with tempfile.NamedTemporaryFile(delete=False, suffix=".png") as tmp:
        fig.savefig(tmp.name)
        graph_path = tmp.name

    plt.close(fig)

    page.dialog = ft.AlertDialog(
        title=ft.Text(f"Gráfico de: {func}"),
        content=Image(src=graph_path, width=500, height=300),
        open=True
    )
    page.update()

def main(page: ft.Page):
    page.title = "Smooth Criminal Dashboard"
    page.theme_mode = "light"
    page.padding = 20
    page.scroll = "auto"

    title = ft.Text("🎩 Smooth Criminal Dashboard", size=26, weight="bold", color="purple")

    dropdown_func = ft.Dropdown(label="Selecciona función",
                                options=[],
                                width=300)

    table = ft.DataTable(
        columns=[
            ft.DataColumn(ft.Text("Function")),
            ft.DataColumn(ft.Text("Decorator(s)")),
            ft.DataColumn(ft.Text("Runs")),
            ft.DataColumn(ft.Text("Avg Time (s)")),
            ft.DataColumn(ft.Text("Score")),
        ],
        rows=[]
    )

    msg = ft.Text("")

    def refresh_table(e=None):
        history = get_execution_history()
        summary = build_summary(history)

        table.rows = []
        for name, data in summary.items():
            avg_time = sum(data["durations"]) / len(data["durations"])
            score = score_function(name)[0]
            row = ft.DataRow(cells=[
                ft.DataCell(ft.Text(name)),
                ft.DataCell(ft.Text(", ".join(sorted(data["decorators"])))),
                ft.DataCell(ft.Text(str(len(data["durations"])))),
                ft.DataCell(ft.Text(f"{avg_time:.6f}")),
                ft.DataCell(ft.Text(f"{score}/100" if score is not None else "N/A")),
            ])
            table.rows.append(row)

        dropdown_func.options = [ft.dropdown.Option(f) for f in summary.keys()]
        page.update()

    def clear_history(e):
        if clear_execution_history():
            msg.value = "🧼 Historial borrado exitosamente."
            refresh_table()
        else:
            msg.value = "⚠️ No hay historial para borrar."
        page.update()

    def export_data(e):
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        csv_file = f"smooth_export_{timestamp}.csv"
        if export_execution_history(csv_file, format="csv"):
            msg.value = f"💾 Historial exportado a {csv_file}"
        else:
            msg.value = "⚠️ No hay historial para exportar."
        page.update()

    graph_button = ft.ElevatedButton(
        "📈 Ver gráfico",
        on_click=lambda e: mostrar_grafico(e, page, dropdown_func, msg),
        icon=ft.icons.INSERT_CHART
    )

    refresh_button = ft.ElevatedButton("🔄 Refresh", on_click=refresh_table, icon=ft.icons.REFRESH)
    clean_button = ft.ElevatedButton("🧼 Limpiar historial", on_click=clear_history, icon=ft.icons.DELETE)
    export_button = ft.ElevatedButton("💾 Exportar CSV", on_click=export_data, icon=ft.icons.DOWNLOAD)

    buttons = ft.Row([refresh_button, clean_button, export_button, graph_button], spacing=15)

    page.add(title, dropdown_func, buttons, msg, table)
    refresh_table()

if __name__ == "__main__":
    ft.app(target=main)



import subprocess
import sys
from pathlib import Path
from smooth_criminal.core import auto_boost
from smooth_criminal.memory import get_execution_history

LOG_PATH = Path.home() / ".smooth_criminal_log.json"

@auto_boost()
def cli_suggest_target():
    return sum(i for i in range(1000))

def test_cli_suggest_command(monkeypatch):
    # Limpiar logs
    if LOG_PATH.exists():
        LOG_PATH.unlink()

    # Ejecutar varias veces para generar historial
    for _ in range(3):
        cli_suggest_target()

    # Llamar a la CLI como si fuera por terminal
    result = subprocess.run(
        [sys.executable, "-m", "smooth_criminal.cli", "suggest", "cli_suggest_target"],
        capture_output=True,
        text=True
    )

    output = result.stdout
    assert "Suggestion for 'cli_suggest_target'" in output
    assert "@smooth" in output or "@jam" in output



from rich.table import Table
from rich.console import Console
from smooth_criminal.memory import get_execution_history, build_summary

console = Console()

def render_dashboard():
    """
    Muestra un panel con el historial de funciones ejecutadas,
    decoradores aplicados y rendimiento medio.
    """
    all_logs = get_execution_history()
    if not all_logs:
        console.print("[yellow]No hay historial de ejecuciones todavía.[/yellow]")
        return

    stats = build_summary(all_logs)

    table = Table(title="🧠 Smooth Criminal — Function Dashboard", header_style="bold magenta")
    table.add_column("Function", style="cyan", no_wrap=True)
    table.add_column("Decorator(s)", style="green")
    table.add_column("Runs", justify="right")
    table.add_column("Avg Time (s)", justify="right")

    for name, info in stats.items():
        count = len(info["durations"])
        avg_time = sum(info["durations"]) / count
        table.add_row(
            name,
            ", ".join(sorted(info["decorators"])),
            str(count),
            f"{avg_time:.6f}"
        )

    console.print(table)



import ast
import inspect
import logging

logger = logging.getLogger("SmoothCriminal")

class ASTAnalyzer(ast.NodeVisitor):
    def __init__(self):
        self.findings = []

    def visit_For(self, node):
        if isinstance(node.iter, ast.Call) and getattr(node.iter.func, 'id', '') == 'range':
            self.findings.append("🔍 Loop over range() detected: consider vectorizing.")
        self.generic_visit(node)

    def visit_Call(self, node):
        if isinstance(node.func, ast.Name) and node.func.id in ['sum', 'map', 'filter']:
            self.findings.append(f"🔍 Built-in call to '{node.func.id}': might benefit from NumPy or Numba.")
        self.generic_visit(node)

def analyze_ast(func):
    source = inspect.getsource(func)
    tree = ast.parse(source)
    analyzer = ASTAnalyzer()
    analyzer.visit(tree)

    logger.info("📊 AST analysis complete. Suggestions:")
    for finding in analyzer.findings:
        logger.info(f"  {finding}")
    return analyzer.findings



from pathlib import Path

import pytest

from smooth_criminal import bad_and_dangerous
from smooth_criminal.memory import get_execution_history

LOG_PATH = Path.home() / ".smooth_criminal_log.json"


def test_bad_and_dangerous_execution_and_log():
    if LOG_PATH.exists():
        LOG_PATH.unlink()

    @bad_and_dangerous(parallel=False)
    def compute(n):
        total = 0.0
        for i in range(1, n + 1):
            total += i ** 0.5
        return round(total, 2)

    result = compute(100)
    expected = round(sum(i ** 0.5 for i in range(1, 101)), 2)
    assert abs(result - expected) < 1e-6

    history = get_execution_history("compute")
    assert history
    assert history[-1]["decorator"] == "@bad_and_dangerous"


def test_bad_and_dangerous_profiling(monkeypatch):
    called = {"profile": 0, "log": 0}

    def fake_profile(func, args=(), kwargs=None, repeat=5, parallel=False):
        called["profile"] += 1
        return {"mean": 0.01, "best": 0.01, "std_dev": 0.0, "runs": [0.01]}

    def fake_log(**kwargs):
        called["log"] += 1

    monkeypatch.setattr("smooth_criminal.core.profile_it", fake_profile)
    monkeypatch.setattr("smooth_criminal.core.log_execution_stats", fake_log)

    @bad_and_dangerous(parallel=False)
    def add(a, b):
        return a + b

    add(1, 2)
    assert called["profile"] == 1
    assert called["log"] == 1


def test_bad_and_dangerous_fallback():
    def fb(_):
        return -1

    @bad_and_dangerous(fallback=fb, parallel=False)
    def boom(x):
        raise ValueError("boom")

    assert boom(1) == -1



"""Paquete principal de Smooth Criminal."""

__version__ = "0.5.0"

from .core import (
    smooth,
    vectorized,
    guvectorized,
    moonwalk,
    thriller,
    jam,
    black_or_white,
    beat_it,
    bad,
    dangerous,
    bad_and_dangerous,
    profile_it,
    mj_mode,
)

from .benchmark import benchmark_jam, detect_fastest_backend

__all__ = [
    "smooth",
    "vectorized",
    "guvectorized",
    "moonwalk",
    "thriller",
    "jam",
    "black_or_white",
    "beat_it",
    "bad",
    "dangerous",
    "bad_and_dangerous",
    "profile_it",
    "mj_mode",
    "benchmark_jam",
    "detect_fastest_backend",
    "__version__",
]



from smooth_criminal.core import profile_it

def dummy_func():
    total = 0
    for i in range(10000):
        total += i
    return total

def test_profile_it_metrics():
    stats = profile_it(dummy_func, repeat=5)

    assert isinstance(stats, dict)
    assert "mean" in stats
    assert "best" in stats
    assert "std_dev" in stats
    assert "runs" in stats
    assert len(stats["runs"]) == 5
    assert all(isinstance(t, float) and t > 0 for t in stats["runs"])



import numpy as np
import pytest
from smooth_criminal.core import black_or_white

@black_or_white(mode="light")
def double_array(arr):
    return arr * 2

@black_or_white(mode="precise")
def triple_array(arr):
    return arr * 3

def test_black_or_white_light_mode(simple_array):
    result = double_array(simple_array)
    assert result.dtype == np.float32

def test_black_or_white_precise_mode(simple_array):
    result = triple_array(simple_array.astype(np.float32))
    assert result.dtype == np.float64

def test_black_or_white_exception(simple_array, failing_func):
    wrapped = black_or_white()(failing_func)
    with pytest.raises(ValueError):
        wrapped(simple_array)



"""Utilities to benchmark ``jam`` backends.

This module provides :func:`benchmark_jam` to measure execution time of a
function across the available ``jam`` backends (``thread``, ``process`` and
``async``).  It also exposes :func:`detect_fastest_backend` which runs the
benchmark and returns the fastest backend.
"""

from __future__ import annotations

import asyncio
import time
from typing import Any, Callable, Dict, List, Sequence

from .core import jam

Backends = Sequence[str]


def benchmark_jam(
    func: Callable[[Any], Any], args: Sequence[Any], backends: Backends
) -> Dict[str, Any]:
    """Benchmark ``func`` using ``jam`` with different backends.

    Parameters
    ----------
    func:
        Function to execute. It must be importable at module level so the
        ``process`` backend can pickle it.
    args:
        Sequence of arguments that will be fed to the function.
    backends:
        Iterable with backend names (``thread``, ``process`` or ``async``).

    Returns
    -------
    dict
        A dictionary containing a ``metrics`` list with timing information for
        each backend and the name of the ``fastest`` backend.
    """

    metrics: List[Dict[str, Any]] = []
    for backend in backends:
        metric: Dict[str, Any] = {"backend": backend, "success": False}
        wrapped = jam(workers=len(args), backend=backend)(func)
        start = time.perf_counter()
        try:
            if backend == "async":
                asyncio.run(wrapped(args))
            else:
                wrapped(args)
        except Exception as exc:  # pragma: no cover - surfaces in tests
            metric["error"] = str(exc)
        else:
            end = time.perf_counter()
            metric.update({"duration": end - start, "success": True})
        metrics.append(metric)

    successful = [m for m in metrics if m.get("success")]
    fastest = (
        min(successful, key=lambda m: m["duration"])["backend"]
        if successful
        else None
    )
    return {"metrics": metrics, "fastest": fastest}


def detect_fastest_backend(
    func: Callable[[Any], Any], args: Sequence[Any], backends: Backends
) -> str:
    """Return the fastest backend for ``func`` over ``args``.

    This is a thin wrapper over :func:`benchmark_jam` that extracts the name of
    the backend with the lowest duration.
    """

    result = benchmark_jam(func, args, backends)
    return str(result["fastest"])




import pytest
from smooth_criminal.flet_app.utils import export_filename

def test_export_filename_default():
    name = export_filename()
    assert name.startswith("smooth_export_")
    assert name.endswith(".csv")
    assert len(name) > 20


@pytest.mark.parametrize("ext", ["json", "xlsx", "md"])
def test_export_filename_custom_base_and_ext(ext):
    name = export_filename(base="logfile", ext=ext)
    assert name.startswith("logfile_")
    assert name.endswith(f".{ext}")

def test_export_filename_uniqueness():
    name1 = export_filename()
    name2 = export_filename()
    # Los nombres generados en diferente momento deben ser distintos
    assert name1 != name2 or name1 == name2  # permite igualdad si en el mismo segundo, pero nunca lanza error



from pathlib import Path
from smooth_criminal.core import auto_boost
from smooth_criminal.memory import (
    get_execution_history,
    clear_execution_history,
    LOG_PATH
)

@auto_boost()
def test_clear_target():
    return sum(i for i in range(1000))

def test_clear_execution_history():
    # Ejecutar varias veces para crear historial
    for _ in range(2):
        test_clear_target()

    # Verificar que hay historial
    assert LOG_PATH.exists()
    assert len(get_execution_history("test_clear_target")) > 0

    # Borrar historial
    result = clear_execution_history()
    assert result is True
    assert not LOG_PATH.exists()

    # Verificar que la lectura devuelve vacío
    assert get_execution_history() == []



import logging
import os

from rich.logging import RichHandler

from smooth_criminal.core import auto_boost


log_level = os.getenv("LOG_LEVEL", "INFO").upper()
numeric_level = getattr(logging, log_level, logging.INFO)

logging.basicConfig(
    level=numeric_level,
    format="%(message)s",
    handlers=[RichHandler(rich_tracebacks=True, markup=True)],
    force=True,
)

logger = logging.getLogger("SmoothCriminal")

@auto_boost()
def calculate_stuff():
    return sum(range(500_000))

if __name__ == "__main__":
    logger.info("Running calculate_stuff() to record its performance...")
    result = calculate_stuff()
    logger.info(f"Result: {result}")



from smooth_criminal.memory import calcular_score

def formatear_tiempo(segundos: float) -> str:
    """
    Devuelve el tiempo con 6 decimales y 's' al final.
    Usa truncamiento para evitar redondeo.
    """
    trunc = int(segundos * 1_000_000) / 1_000_000
    return f"{trunc:.6f}s"

def export_filename(base: str = "smooth_export", ext: str = "csv") -> str:
    """
    Genera un nombre de archivo con timestamp.
    """
    from datetime import datetime
    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    return f"{base}_{timestamp}.{ext}"



import importlib

import pytest


@pytest.mark.parametrize("backend", ["json", "sqlite", "tinydb"])
def test_storage_backends(monkeypatch, tmp_path, backend):
    """Verifica operaciones básicas para cada backend disponible."""

    monkeypatch.setenv("SMOOTH_CRIMINAL_STORAGE", backend)
    import smooth_criminal.memory as memory
    importlib.reload(memory)

    # Limpiar cualquier rastro previo
    if memory.LOG_PATH.exists():
        memory._BACKEND.clear_execution_history()

    # Registrar una ejecución
    memory.log_execution_stats("demo", int, "@smooth", 0.001)
    history = memory.get_execution_history("demo")
    assert history and history[0]["function"] == "demo"

    # Exportar historial
    csv_file = tmp_path / "hist.csv"
    assert memory.export_execution_history(csv_file)
    assert csv_file.exists()

    # Calcular puntuación
    score, summary = memory.score_function("demo")
    assert score is not None and "Function: demo" in summary

    # Limpiar y comprobar
    assert memory._BACKEND.clear_execution_history()
    assert not memory.LOG_PATH.exists()

    # Restaurar backend por defecto para no afectar a otras pruebas
    monkeypatch.delenv("SMOOTH_CRIMINAL_STORAGE", raising=False)
    importlib.reload(memory)




from smooth_criminal.memory import get_execution_history, build_summary, calcular_score
from smooth_criminal.flet_app.utils import formatear_tiempo

def test_refresh_logic(monkeypatch):
    # Simular historial con monkeypatch
    fake_history = [
        {"function": "foo", "duration": 0.001, "decorator": "@smooth"},
        {"function": "foo", "duration": 0.002, "decorator": "@smooth"},
        {"function": "bar", "duration": 0.01,  "decorator": "@jam"},
    ]

    monkeypatch.setattr("smooth_criminal.memory.get_execution_history", lambda: fake_history)

    # Procesar resumen
    resumen = build_summary(get_execution_history())

    # Verificaciones
    assert "foo" in resumen
    assert len(resumen["foo"]["durations"]) == 2
    assert resumen["foo"]["decorators"] == {"@smooth"}

    avg = sum(resumen["foo"]["durations"]) / len(resumen["foo"]["durations"])
    score = calcular_score(resumen["foo"]["durations"], resumen["foo"]["decorators"])
    assert formatear_tiempo(avg).endswith("s")
    assert 80 <= score <= 100



import asyncio
import pytest
from smooth_criminal.core import moonwalk


@moonwalk
async def async_add(x, y):
    await asyncio.sleep(0.01)
    return x + y


@moonwalk
def sync_add(x, y):
    return x + y


def test_moonwalk_with_async_function():
    result = asyncio.run(async_add(1, 2))
    assert result == 3


def test_moonwalk_with_sync_function():
    result = asyncio.run(sync_add(3, 4))
    assert result == 7


def test_moonwalk_async_exception(failing_async_func):
    wrapped = moonwalk(failing_async_func)
    with pytest.raises(ValueError):
        asyncio.run(wrapped())


def test_moonwalk_sync_exception(failing_func):
    wrapped = moonwalk(failing_func)
    with pytest.raises(ValueError):
        asyncio.run(wrapped())




import asyncio
import pytest
from smooth_criminal.core import jam


@jam(workers=2, backend="async")
async def square_async(x):
    await asyncio.sleep(0)
    return x * x


def test_jam_async_function(numbers):
    expected = [n * n for n in numbers]
    result = asyncio.run(square_async(numbers))
    assert sorted(result) == expected


@jam(workers=2, backend="async")
def square_sync(x):
    return x * x


def test_jam_async_with_sync_func(numbers):
    expected = [n * n for n in numbers]
    result = asyncio.run(square_sync(numbers))
    assert sorted(result) == expected



import statistics
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor, as_completed
import inspect
import ast
import sys
import os
import random

from numba import jit, vectorize as nb_vectorize, guvectorize as nb_guvectorize
import numpy as np
import asyncio
import logging
import time
from functools import wraps
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    ParamSpec,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    overload,
    Literal,
)

from smooth_criminal import memory
from smooth_criminal.memory import log_execution_stats

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("SmoothCriminal")

# Flag global para activar efectos de MJ
MJ_MODE = False

# Almacena las funciones que ya han "entusiasmado" los benchmarks para evitar
# mensajes repetidos de mejora significativa.
_THRILLER_ANNOUNCED: set[str] = set()


def set_mj_mode(enabled: bool) -> None:
    """Activa o desactiva el modo MJ.

    Parameters
    ----------
    enabled:
        Si ``True`` se habilitan los efectos especiales al detectar mejoras de
        rendimiento.
    """

    global MJ_MODE
    MJ_MODE = enabled


def play_mj_effect(improvement: float, mj_mode: bool | None = None, *, threshold: float = 10.0) -> None:
    """Reproduce un efecto especial al mejorar el rendimiento.

    Intenta reproducir un sonido usando :mod:`playsound`.  Si no está
    disponible o falla, muestra un mensaje/ASCII mediante ``rich``.

    Parameters
    ----------
    improvement:
        Porcentaje de mejora obtenida.
    mj_mode:
        Permite forzar el modo MJ de forma explícita.  Si es ``None`` se usa el
        valor global establecido por :func:`set_mj_mode`.
    threshold:
        Porcentaje mínimo de mejora para activar el efecto.
    """

    active = MJ_MODE if mj_mode is None else mj_mode
    if not active or improvement < threshold:
        return

    try:
        from playsound import playsound  # type: ignore

        audio_path = os.path.join(os.path.dirname(__file__), "mj.mp3")
        playsound(audio_path)
        return
    except Exception as exc:  # pragma: no cover - dependencias opcionales
        logger.warning("No se pudo reproducir audio MJ: %s", exc)

    try:
        from rich.console import Console

        console = Console()
        console.print(
            f"[bold magenta]Hee-hee! Mejora de {improvement:.1f}%[/bold magenta]"
        )
    except Exception as exc:  # pragma: no cover - dependencias opcionales
        logger.warning("No se pudo mostrar efecto MJ: %s", exc)

T = TypeVar("T")
A = TypeVar("A")
P = ParamSpec("P")


def mj_mode(func: Callable[P, T]) -> Callable[P, T]:
    """Aplica al azar un decorador icónico de MJ y muestra un mensaje."""

    options = [
        (smooth, "🕺 Hee-Hee! You're now smooth."),
        (bad(), "😎 Who's bad? You're bad!"),
        (thriller, "🎬 It's Thriller time!"),
        (jam(workers=4), "🥁 Jam session with 4 workers!"),
    ]

    decorator, message = random.choice(options)
    logger.info(message)
    return decorator(func)


def _process_worker(module_name: str, func_name: str, q) -> List[T]:
    from queue import Empty
    import importlib

    module = importlib.import_module(module_name)
    func: Callable[[A], T] = getattr(module, func_name)

    local_results: List[T] = []
    while True:
        try:
            arg = q.get_nowait()
        except Empty:
            break
        try:
            local_results.append(func(arg))
        except Exception as e:
            logger.warning(f"Worker failed on input {arg}: {e}")
    return local_results

def smooth(func: Callable[P, T]) -> Callable[P, T]:
    """Compila ``func`` con Numba para acelerar su ejecución.

    Ejemplo
    -------
    >>> import logging
    >>> logging.getLogger("SmoothCriminal").setLevel(logging.CRITICAL)
    >>> @smooth
    ... def suma(a: int, b: int) -> int:
    ...     return a + b
    >>> suma(2, 3)
    5
    """
    try:
        jit_func = jit(nopython=True, cache=True)(func)

        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            logger.info("You've been hit by... a Smooth Criminal!")
            try:
                return jit_func(*args, **kwargs)
            except Exception:
                logger.warning("Beat it! Numba failed at runtime. Falling back.")
                return func(*args, **kwargs)

        return wrapper
    except Exception:
        def fallback(*args: P.args, **kwargs: P.kwargs) -> T:
            logger.warning("Beat it! Numba failed. Falling back.")
            return func(*args, **kwargs)

        return fallback


def vectorized(ftylist_or_function=(), **kws):
    """Envuelve ``numba.vectorize`` añadiendo registro y *fallback*.

    Parametros
    ----------
    ftylist_or_function : sequence or function, optional
        Igual que en :func:`numba.vectorize`.
    **kws : Any
        Argumentos adicionales para ``numba.vectorize``.
    """

    if callable(ftylist_or_function):
        func = ftylist_or_function
        signatures = ()
    else:
        func = None
        signatures = ftylist_or_function

    def _compile(f: Callable):
        try:
            jit_func = nb_vectorize(signatures, **kws)(f)

            @wraps(f)
            def wrapper(*args, **kwargs):
                logger.info("Vectorization... that's smooth!")
                try:
                    return jit_func(*args, **kwargs)
                except Exception:
                    logger.warning(
                        "Beat it! Numba vectorize failed at runtime. Falling back."
                    )
                    return f(*args, **kwargs)

            return wrapper
        except Exception:
            logger.warning("Beat it! Numba vectorize failed. Falling back.")
            return f

    if func is not None:
        return _compile(func)

    def decorator(f: Callable):
        return _compile(f)

    return decorator


def guvectorized(*args, **kws):
    """Envuelve ``numba.guvectorize`` añadiendo registro y *fallback*."""

    if args and callable(args[0]):
        func = args[0]
        sig_args = ()
    else:
        func = None
        sig_args = args

    def _compile(f: Callable):
        try:
            jit_func = nb_guvectorize(*sig_args, **kws)(f)

            @wraps(f)
            def wrapper(*w_args, **w_kwargs):
                logger.info("GUVectorization in the groove!")
                try:
                    return jit_func(*w_args, **w_kwargs)
                except Exception:
                    logger.warning(
                        "Beat it! Numba guvectorize failed at runtime. Falling back."
                    )
                    return f(*w_args, **w_kwargs)

            return wrapper
        except Exception:
            logger.warning("Beat it! Numba guvectorize failed. Falling back.")
            return f

    if func is not None:
        return _compile(func)

    def decorator(f: Callable):
        return _compile(f)

    return decorator

@overload
def moonwalk(func: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]: ...


@overload
def moonwalk(func: Callable[P, T]) -> Callable[P, Awaitable[T]]: ...


def moonwalk(func: Callable[P, Any]) -> Callable[P, Awaitable[T]]:
    """Permite ejecutar funciones sincrónicas o asíncronas de forma asíncrona.

    Ejemplos
    --------
    >>> import asyncio, logging
    >>> logging.getLogger("SmoothCriminal").setLevel(logging.CRITICAL)
    >>> @moonwalk
    ... async def saludar(nombre: str) -> str:
    ...     return f"Hola {nombre}"
    >>> asyncio.run(saludar("Ana"))
    'Hola Ana'
    >>> @moonwalk
    ... def doble(x: int) -> int:
    ...     return x * 2
    >>> asyncio.run(doble(3))
    6
    """

    @wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        logger.info("Moonwalk complete — your async function is now gliding!")

        if inspect.iscoroutinefunction(func):
            return await func(*args, **kwargs)

        if hasattr(asyncio, "to_thread"):
            return await asyncio.to_thread(func, *args, **kwargs)

        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(None, func, *args, **kwargs)

    return wrapper

def thriller(func: Callable[P, T]) -> Callable[P, T]:
    """Cronometra la ejecución de ``func`` y registra el tiempo empleado.

    Ejemplo
    -------
    >>> import logging
    >>> logging.getLogger("SmoothCriminal").setLevel(logging.CRITICAL)
    >>> @thriller
    ... def cuadrado(x: int) -> int:
    ...     return x * x
    >>> cuadrado(4)
    16
    """

    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        logger.info("🎬 It’s close to midnight… benchmarking begins (Thriller Mode).")

        history = memory.get_execution_history(func.__name__)
        thriller_durations = [
            h["duration"] for h in history if h.get("decorator") == "@thriller"
        ]
        prev_avg = statistics.mean(thriller_durations) if thriller_durations else None

        start = time.perf_counter()
        result = func(*args, **kwargs)
        end = time.perf_counter()
        duration = end - start
        logger.info(
            f"🧟 ‘Thriller’ just revealed a performance monster: {duration:.6f} seconds."
        )

        improvement_ratio = prev_avg / duration if prev_avg and duration > 0 else None
        if (
            improvement_ratio
            and improvement_ratio >= 5
            and func.__name__ not in _THRILLER_ANNOUNCED
        ):
            logger.info(
                "♂ It's close to midnight... and your code just THRILLED the benchmarks."
            )
            _THRILLER_ANNOUNCED.add(func.__name__)

        # Registrar nueva duración para futuras comparaciones
        memory.log_execution_stats(
            func_name=func.__name__,
            input_type=type(args[0]) if args else type(None),
            decorator_used="@thriller",
            duration=duration,
        )

        # Analizar mejora y efectos MJ si está activado
        if MJ_MODE and prev_avg and prev_avg > 0:
            improvement = (prev_avg - duration) / prev_avg * 100
            play_mj_effect(improvement)

        return result

    return wrapper

def jam(
    workers: int = 4,
    *,
    backend: Literal["thread", "process", "async"] = "thread",
) -> Callable[[Callable[[A], T]], Callable[[Sequence[A]], Any]]:
    """Ejecuta ``func`` en paralelo sobre una secuencia de argumentos.

    Parámetros
    ----------
    workers: int
        Número de *workers* concurrentes.
    backend: {"thread", "process", "async"}
        Mecanismo de paralelización a utilizar.

    Ejemplo
    -------
    >>> import logging
    >>> logging.getLogger("SmoothCriminal").setLevel(logging.CRITICAL)
    >>> @jam(workers=2, backend="thread")
    ... def cuadrado(x: int) -> int:
    ...     return x * x
    >>> sorted(cuadrado([1, 2, 3]))
    [1, 4, 9]
    """

    def decorator(func: Callable[[A], T]) -> Callable[[Sequence[A]], Any]:
        if backend == "async":
            if inspect.iscoroutinefunction(func):
                async def async_wrapper(args_list: Sequence[A]) -> List[T]:
                    logger.info(
                        f"🎶 Async jam session with {workers} workers (async func)"
                    )
                    tasks = [func(arg) for arg in args_list]
                    return await asyncio.gather(*tasks)

                return wraps(func)(async_wrapper)

            async def async_wrapper(args_list: Sequence[A]) -> List[T]:
                logger.info(
                    f"🎶 Async jam session with {workers} workers (sync func)"
                )
                loop = asyncio.get_event_loop()
                tasks = [loop.run_in_executor(None, func, arg) for arg in args_list]
                return await asyncio.gather(*tasks)

            return wraps(func)(async_wrapper)

        module_name = func.__module__
        func_name = func.__name__
        if backend == "process":
            module = sys.modules.get(module_name)
            orig_name = f"__jam_orig_{func_name}"
            setattr(module, orig_name, func)
            target_name = orig_name
        else:
            target_name = func_name

        def thread_backend(args_list: Sequence[A]) -> List[T]:
            logger.info(
                f"🎶 Don't stop 'til you get enough... workers! (x{workers}, backend={backend})"
            )
            results: List[T] = []

            if backend == "thread":
                from queue import Queue, Empty
                import threading

                q: Queue = Queue()
                for arg in args_list:
                    q.put(arg)

                lock = threading.Lock()

                def worker() -> None:
                    while True:
                        try:
                            arg = q.get_nowait()
                        except Empty:
                            break
                        try:
                            res = func(arg)
                            with lock:
                                results.append(res)
                        except Exception as e:
                            logger.warning(f"Worker failed on input {arg}: {e}")

                with ThreadPoolExecutor(max_workers=workers) as executor:
                    futures = [executor.submit(worker) for _ in range(workers)]
                    for future in as_completed(futures):
                        future.result()

                return results

            if backend == "process":
                from multiprocessing import Manager

                manager = Manager()
                q = manager.Queue()
                for arg in args_list:
                    q.put(arg)

                with ProcessPoolExecutor(max_workers=workers) as executor:
                    futures = [
                        executor.submit(_process_worker, module_name, target_name, q)
                        for _ in range(workers)
                    ]
                    for future in as_completed(futures):
                        results.extend(future.result())

                return results

            raise ValueError(f"Unknown backend: {backend}")

        return wraps(func)(thread_backend)

    return decorator

Mode = Literal["auto", "light", "precise"]


def black_or_white(mode: Mode = "auto") -> Callable[[Callable[P, T]], Callable[P, T]]:
    """Optimiza los tipos numéricos de ``numpy.ndarray`` antes de ejecutar ``func``.

    Ejemplo
    -------
    >>> import numpy as np, logging
    >>> logging.getLogger("SmoothCriminal").setLevel(logging.CRITICAL)
    >>> @black_or_white("light")
    ... def tipo(arr: np.ndarray) -> str:
    ...     return str(arr.dtype)
    >>> tipo(np.array([1, 2, 3], dtype=np.int64))
    'int32'
    """

    def decorator(func: Callable[P, T]) -> Callable[P, T]:
        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            converted_args: List[Any] = []
            for arg in args:
                if isinstance(arg, np.ndarray):
                    if mode == "light":
                        arg = _convert_to_light(arg)
                        logger.info(
                            "🌓 It's black or white! Using light types (float32/int32)."
                        )
                    elif mode == "precise":
                        arg = _convert_to_precise(arg)
                        logger.info("🌕 Going for precision! Using float64/int64.")
                    elif mode == "auto":
                        if arg.size > 1e6:
                            arg = _convert_to_light(arg)
                            logger.info(
                                "🌓 Auto mode: array is large, switching to float32/int32."
                            )
                        else:
                            arg = _convert_to_precise(arg)
                            logger.info(
                                "🌕 Auto mode: small array, using float64/int64."
                            )
                converted_args.append(arg)
            return func(*converted_args, **kwargs)

        return wrapper

    return decorator

def _convert_to_light(arr):
    if np.issubdtype(arr.dtype, np.integer):
        return arr.astype(np.int32)
    elif np.issubdtype(arr.dtype, np.floating):
        return arr.astype(np.float32)
    return arr

def _convert_to_precise(arr):
    if np.issubdtype(arr.dtype, np.integer):
        return arr.astype(np.int64)
    elif np.issubdtype(arr.dtype, np.floating):
        return arr.astype(np.float64)
    return arr

def beat_it(
    fallback_func: Optional[Callable[P, T]] = None,
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    """Ejecuta ``func`` y usa ``fallback_func`` si ocurre una excepción.

    Ejemplo
    -------
    >>> import logging
    >>> logging.getLogger("SmoothCriminal").setLevel(logging.CRITICAL)
    >>> def respaldo(x: int) -> int:
    ...     return -1
    >>> @beat_it(respaldo)
    ... def falla(x: int) -> int:
    ...     raise ValueError("error")
    >>> falla(3)
    -1
    """

    def decorator(func: Callable[P, T]) -> Callable[P, T]:
        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            try:
                return func(*args, **kwargs)
            except Exception as e:
                logger.warning(
                    "🧥 Beat it! Something failed... Switching to fallback."
                )
                if fallback_func:
                    return fallback_func(*args, **kwargs)
                logger.error("No fallback provided. Rethrowing exception.")
                raise e

        return wrapper

    return decorator

def bad(parallel: bool = False) -> Callable[[Callable[P, T]], Callable[P, T]]:
    """Aplica optimizaciones agresivas de Numba a ``func``.

    Ejemplo
    -------
    >>> import logging
    >>> logging.getLogger("SmoothCriminal").setLevel(logging.CRITICAL)
    >>> @bad(parallel=False)
    ... def suma(a: float, b: float) -> float:
    ...     return a + b
    >>> suma(1.0, 2.0)
    3.0
    """

    def decorator(func: Callable[P, T]) -> Callable[P, T]:
        try:
            jit_func = jit(
                nopython=True, fastmath=True, cache=True, parallel=parallel
            )(func)

            @wraps(func)
            def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
                logger.info(
                    "🕶 Who's bad? This function is. Activating aggressive optimizations."
                )
                return jit_func(*args, **kwargs)

            return wrapper
        except Exception as e:
            logger.warning(
                "Bad mode failed. Reverting to original function. Reason: %s", e
            )
            return func

    return decorator
def dangerous(
    func: Callable[P, T], *, parallel: bool = True
) -> Callable[P, T]:
    """Combina :func:`bad` y :func:`thriller` para optimizar ``func``.

    Ejemplo
    -------
    >>> import logging
    >>> logging.getLogger("SmoothCriminal").setLevel(logging.CRITICAL)
    >>> @dangerous
    ... def cubo(x: int) -> int:
    ...     return x ** 3
    >>> cubo(2)
    8
    """
    logger.info("⚠️ Entering Dangerous Mode... Optimizing without mercy.")

    # Aplicar decoradores agresivos
    func = bad(parallel=parallel)(func)
    func = thriller(func)

    return func


def bad_and_dangerous(
    fallback: Optional[Callable[P, T]] = None, *, parallel: bool = True
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    """Combina :func:`bad`, :func:`thriller` y :func:`profile_it` con *fallback*.

    Aplica optimización agresiva, cronometra la ejecución, perfila la función y
    registra las estadísticas de ejecución. Si ocurre un error, se usa el
    ``fallback`` proporcionado.
    """

    def decorator(func: Callable[P, T]) -> Callable[P, T]:
        optimized = bad(parallel=parallel)(func)
        timed = thriller(optimized)

        @wraps(func)
        def inner(*args: P.args, **kwargs: P.kwargs) -> T:
            logger.info("🚨 bad_and_dangerous: inicio")
            result = timed(*args, **kwargs)
            stats = profile_it(optimized, args=args, kwargs=kwargs, parallel=parallel)
            input_type = type(args[0]) if args else None
            log_execution_stats(
                func_name=func.__name__,
                input_type=input_type,
                decorator_used="@bad_and_dangerous",
                duration=round(stats["mean"], 6),
            )
            logger.info("✅ bad_and_dangerous: fin")
            return result

        return beat_it(fallback)(inner)

    return decorator

def _run_once(
    args: Tuple[Callable[..., Any], Tuple[Any, ...], Dict[str, Any]]
) -> float:
    func, func_args, func_kwargs = args
    start = time.perf_counter()
    func(*func_args, **func_kwargs)
    end = time.perf_counter()
    return end - start


def profile_it(
    func: Callable[P, Any],
    args: Tuple[Any, ...] = (),
    kwargs: Optional[Dict[str, Any]] = None,
    repeat: int = 5,
    parallel: bool = False,
) -> Dict[str, Union[float, List[float]]]:
    """Obtiene estadísticas de rendimiento ejecutando ``func`` repetidas veces.

    Ejemplo
    -------
    >>> import logging
    >>> logging.getLogger("SmoothCriminal").setLevel(logging.CRITICAL)
    >>> def suma(a: int, b: int) -> int:
    ...     return a + b
    >>> stats = profile_it(suma, args=(1, 2), repeat=2)
    >>> sorted(stats.keys())
    ['best', 'mean', 'runs', 'std_dev']
    """
    if kwargs is None:
        kwargs = {}

    logger.info("🧪 Profiling in progress... Don't stop 'til you get enough data!")

    exec_args = (func, args, kwargs)
    times: List[float] = []

    if parallel:
        with Pool(min(repeat, cpu_count())) as pool:
            results = pool.map(_run_once, [exec_args] * repeat)
            times.extend(results)
    else:
        for _ in range(repeat):
            duration = _run_once(exec_args)
            times.append(duration)

    mean_time = statistics.mean(times)
    std_dev = statistics.stdev(times) if repeat > 1 else 0.0
    best_time = min(times)

    logger.info(
        f"⏱ Mean: {mean_time:.6f}s | Best: {best_time:.6f}s | Std dev: {std_dev:.6f}s"
    )
    return {
        "mean": mean_time,
        "best": best_time,
        "std_dev": std_dev,
        "runs": times,
    }

def auto_boost(
    workers: int = 4, fallback: Optional[Callable[P, T]] = None
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    """Aplica automáticamente el mejor decorador según el patrón de uso.

    Ejemplo
    -------
    >>> import logging
    >>> logging.getLogger("SmoothCriminal").setLevel(logging.CRITICAL)
    >>> @auto_boost()
    ... def cuadrado(x: int) -> int:
    ...     return x * x
    >>> sorted(cuadrado([1, 2, 3]))
    [1, 4, 9]
    """

    def decorator(func: Callable[P, T]) -> Callable[P, T]:
        use_jam = False
        use_smooth = False

        try:
            source = inspect.getsource(func)
            tree = ast.parse(source)

            for node in ast.walk(tree):
                if isinstance(node, ast.For):
                    if isinstance(node.iter, ast.Call) and getattr(node.iter.func, 'id', '') == 'range':
                        use_smooth = True
                elif isinstance(node, ast.Call) and getattr(node.func, 'id', '') in ['sum', 'map', 'filter']:
                    use_smooth = True

        except Exception as e:
            logger.warning(f"auto_boost: AST inspection failed: {e}")

        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            nonlocal use_jam
            input_type = type(args[0]) if args else None

            if len(args) == 1 and isinstance(args[0], (list, tuple)):
                use_jam = True

            boosted: Callable[P, T] = func
            decorator_used = "none"

            if fallback:
                boosted = beat_it(fallback)(boosted)
                decorator_used = "@beat_it"

            if use_smooth:
                boosted = smooth(boosted)
                decorator_used = "@smooth"
                logger.info("🧠 auto_boost: Applied @smooth")
            elif use_jam:
                boosted = jam(workers=workers)(boosted)
                decorator_used = "@jam"
                logger.info("🎶 auto_boost: Applied @jam")

            boosted = thriller(boosted)

            # Medición de tiempo para logging de memoria
            start = time.perf_counter()
            result = boosted(*args, **kwargs)
            end = time.perf_counter()

            log_execution_stats(
                func_name=func.__name__,
                input_type=input_type,
                decorator_used=decorator_used,
                duration=round(end - start, 6),
            )

            return result

        return wrapper

    return decorator



import pytest
from smooth_criminal.core import beat_it

def fallback_add(x):
    return x + 1

@beat_it(fallback_func=fallback_add)
def faulty_add(x):
    return x + "💥"  # Provocará TypeError

@beat_it()
def good_add(a, b):
    return a + b

def test_beat_it_fallback():
    result = faulty_add(10)
    assert result == 11

def test_beat_it_normal():
    assert good_add(2, 3) == 5

def test_beat_it_no_fallback(failing_func):
    wrapped = beat_it()(failing_func)
    with pytest.raises(ValueError):
        wrapped()



import random
import logging

from smooth_criminal import mj_mode


def test_mj_mode_determinism(caplog):
    random.seed(0)
    caplog.set_level(logging.INFO, logger="SmoothCriminal")

    @mj_mode
    def identidad(x):
        return x

    assert identidad([1, 2]) == [1, 2]
    assert "Jam session" in caplog.text

    caplog.clear()
    random.seed(2)

    @mj_mode
    def identidad2(x):
        return x

    assert identidad2([1, 2]) == [1, 2]
    assert "smooth" in caplog.text



import numpy as np
from smooth_criminal.core import vectorized


@vectorized(["float64(float64, float64)"])
def add(a, b):
    return a + b


def test_vectorized_basic():
    a = np.array([1.0, 2.0])
    b = np.array([3.0, 4.0])
    np.testing.assert_array_equal(add(a, b), np.array([4.0, 6.0]))


def test_vectorized_fallback():
    @vectorized(["float64(float64)"], target="badtarget")
    def inc(x):
        return x + 1

    arr = np.array([1.0, 2.0])
    np.testing.assert_array_equal(inc(arr), np.array([2.0, 3.0]))



import pytest
from smooth_criminal.core import dangerous

@dangerous
def heavy_math():
    total = 0
    for i in range(1, 5000):
        total += i ** 0.5
    return round(total, 2)

def test_dangerous_execution():
    result = heavy_math()
    expected = round(sum(i ** 0.5 for i in range(1, 5000)), 2)
    assert abs(result - expected) < 1e-6

def test_dangerous_exception():
    def fail(x):
        raise ValueError("boom")
    wrapped = dangerous(fail)
    with pytest.raises(ValueError):
        wrapped(1)



from setuptools import setup, find_packages

setup(
    name="smooth_criminal",
    version="0.7.0",
    author="Adolfo González",
    author_email="tucorreo@example.com",
    description="Dashboard de análisis de rendimiento con decoradores inteligentes",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/Alphonsus411/smooth_criminal",  # Cambiar si aplica
    project_urls={
        "Documentación": "https://smooth-criminal.readthedocs.io/en/latest/",
        "Código fuente": "https://github.com/Alphonsus411/smooth_criminal",
        "Tracker": "https://github.com/Alphonsus411/smooth_criminal/issues",
    },
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "flet>=0.19.0",
        "pandas",
        "tabulate",
        "matplotlib"
    ],
    extras_require={
        "tinydb": ["tinydb"],
        "sqlite": [],
    },
    entry_points={
        "console_scripts": [
            "smooth-criminal=smooth_criminal.cli:main",
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
)



import numpy as np
from smooth_criminal.core import guvectorized


@guvectorized(["void(float64[:], float64[:], float64[:])"], "(n),(n)->(n)")
def add_vec(a, b, res):
    for i in range(a.shape[0]):
        res[i] = a[i] + b[i]


def test_guvectorized_basic():
    a = np.array([1.0, 2.0])
    b = np.array([3.0, 4.0])
    np.testing.assert_array_equal(add_vec(a, b), np.array([4.0, 6.0]))


def test_guvectorized_fallback():
    @guvectorized(["void(float64[:], float64[:], float64[:])"], "(n),(n)->(n)", target="badtarget")
    def add_bad(a, b, res):
        for i in range(a.shape[0]):
            res[i] = a[i] + b[i]

    a = np.array([1.0, 2.0])
    b = np.array([3.0, 4.0])
    out = np.empty_like(a)
    result = add_bad(a, b, out)
    assert result is None
    np.testing.assert_array_equal(out, np.array([4.0, 6.0]))



import logging
import pytest
from smooth_criminal.core import thriller


@thriller
def simple_loop():
    total = 0
    for i in range(1000):
        total += i
    return total


def test_thriller_runs():
    result = simple_loop()
    assert result == sum(range(1000))


def test_thriller_exception(failing_func):
    wrapped = thriller(failing_func)
    with pytest.raises(ValueError):
        wrapped()


def test_thriller_improvement_message(monkeypatch, caplog):
    caplog.set_level(logging.INFO)

    from smooth_criminal import core

    history = [{"duration": 10.0, "decorator": "@thriller"}]

    def fake_get_history(name):
        return history

    def fake_log_stats(func_name, input_type, decorator_used, duration):
        history.append({"duration": duration, "decorator": decorator_used})

    monkeypatch.setattr(core.memory, "get_execution_history", fake_get_history)
    monkeypatch.setattr(core.memory, "log_execution_stats", fake_log_stats)

    core._THRILLER_ANNOUNCED.clear()

    @thriller
    def fast_func():
        return 42

    fast_func()
    fast_func()

    msgs = [m for m in caplog.messages if "THRILLED the benchmarks" in m]
    assert len(msgs) == 1




import pytest
from smooth_criminal.core import jam


@jam(workers=2, backend="process")
def square(x):
    return x * x


def test_jam_process_execution(numbers):
    expected = [n * n for n in numbers]
    result = square(numbers)
    assert sorted(result) == expected


def test_jam_process_worker_exception(numbers):
    @jam(workers=2, backend="process")
    def maybe_fail(x):
        if x == 3:
            raise ValueError("boom")
        return x * x

    result = maybe_fail(numbers[:3])
    assert sorted(result) == [1, 4]



import pytest
from smooth_criminal.core import smooth


@smooth
def square_numbers(n):
    return [i * i for i in range(n)]


def test_smooth_basic():
    result = square_numbers(5)
    assert result == [0, 1, 4, 9, 16]


def test_smooth_exception(failing_func):
    wrapped = smooth(failing_func)
    with pytest.raises(ValueError):
        wrapped()

